from pathlib import Path
from typing import Callable, Dict, List, Tuple

from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.utils import StrEnum

from webviz_subsurface._models import GruptreeModel
from webviz_subsurface._providers import (
    EnsembleSummaryProvider,
    EnsembleSummaryProviderFactory,
    Frequency,
)

from ._utils import EnsembleGroupTreeData
from ._views._group_tree_view import GroupTreeView, GroupTreeViewElement


class GroupTree(WebvizPluginABC):
    """This plugin vizualizes the network tree and displays pressures,
    rates and other network related information.

    ---
    * **`ensembles`:** Which ensembles in `shared_settings` to include.
    * **`gruptree_file`:** `.csv` with gruptree information.
    * **`time_index`:** Frequency for the data sampling.
    ---

    **Summary data**

    This plugin needs the following summary vectors to be exported:
    * FOPR, FWPR, FOPR, FWIR and FGIR
    * GPR for all group nodes in the network
    * GOPR, GWPR and GGPR for all group nodes in the production network \
    (GOPRNB etc for BRANPROP trees)
    * GGIR and/or GWIR for all group nodes in the injection network
    * WSTAT, WTHP, WBHP, WMCTL for all wells
    * WOPR, WWPR, WGPR for all producers
    * WWIR and/or WGIR for all injectors

    **GRUPTREE input**

    `gruptree_file` is a path to a file stored per realization (e.g. in \
    `share/results/tables/gruptree.csv"`).
    The `gruptree_file` file can be dumped to disk per realization by the `ECL2CSV` forward
    model with subcommand `gruptree`. The forward model uses `ecl2df` to export a table
    representation of the Eclipse network:
    [Link to ecl2csv gruptree documentation.](https://equinor.github.io/ecl2df/usage/gruptree.html).

    **time_index**

    This is the sampling interval of the summary data. It is `yearly` by default, but can be set
    to `monthly` if needed.
    """

    class Ids(StrEnum):
        GROUPTREE_VIEW = "group-tree-view"

    def __init__(
        self,
        webviz_settings: WebvizSettings,
        ensembles: list,
        gruptree_file: str = "share/results/tables/gruptree.csv",
        rel_file_pattern: str = "share/results/unsmry/*.arrow",
        time_index: str = "yearly",
    ) -> None:
        super().__init__(stretch=True)

        self._ensembles = ensembles
        self._gruptree_file = gruptree_file

        if ensembles is None:
            raise ValueError('Incorrect argument, must provide "ensembles"')

        sampling = Frequency(time_index)

        self._ensemble_paths: Dict[str, Path] = {
            ensemble_name: webviz_settings.shared_settings["scratch_ensembles"][
                ensemble_name
            ]
            for ensemble_name in ensembles
        }

        provider_factory = EnsembleSummaryProviderFactory.instance()

        self._group_tree_data: Dict[str, EnsembleGroupTreeData] = {}

        sampling = Frequency(time_index)
        for ens_name, ens_path in self._ensemble_paths.items():
            provider: EnsembleSummaryProvider = (
                provider_factory.create_from_arrow_unsmry_presampled(
                    str(ens_path), rel_file_pattern, sampling
                )
            )
            self._group_tree_data[ens_name] = EnsembleGroupTreeData(
                provider, GruptreeModel(ens_name, ens_path, gruptree_file)
            )

        self.add_view(
            GroupTreeView(self._group_tree_data),
            self.Ids.GROUPTREE_VIEW,
        )

    def add_webvizstore(self) -> List[Tuple[Callable, List[Dict]]]:
        return [
            ens_grouptree_data.webviz_store
            for _, ens_grouptree_data in self._group_tree_data.items()
        ]

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.view(self.Ids.GROUPTREE_VIEW)
                .settings_group(GroupTreeView.Ids.CONTROLS)
                .get_unique_id(),
                "content": "Menu for selecting ensemble and tree mode.",
            },
            {
                "id": self.view(self.Ids.GROUPTREE_VIEW)
                .settings_group(GroupTreeView.Ids.OPTIONS)
                .get_unique_id(),
                "content": "Menu for statistical options or realization.",
            },
            {
                "id": self.view(self.Ids.GROUPTREE_VIEW)
                .settings_group(GroupTreeView.Ids.FILTERS)
                .get_unique_id(),
                "content": "Menu for filtering options.",
            },
            {
                "id": self.view(self.Ids.GROUPTREE_VIEW)
                .view_element(GroupTreeView.Ids.VIEW_ELEMENT)
                .component_unique_id(GroupTreeViewElement.Ids.COMPONENT),
                "content": "Vizualisation of network tree.",
            },
        ]



from pathlib import Path
from typing import Dict, List, Union

import numpy as np
import pandas as pd
import webviz_core_components as wcc
from dash import Dash, Input, Output, html
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.common_cache import CACHE

from webviz_subsurface._providers import EnsembleSummaryProvider

from .._utils.fanchart_plotting import (
    FanchartData,
    FreeLineData,
    LowHighData,
    MinMaxData,
    get_fanchart_traces,
)
from .._utils.unique_theming import unique_colors
from ._simulation_time_series._utils.create_provider_set_from_paths import (
    create_lazy_provider_set_from_paths,
)


class BhpQc(WebvizPluginABC):
    """QC simulated bottom hole pressures (BHP) from reservoir simulations.

    Can be used to check if your simulated BHPs are in a realistic range.
    E.g. check if your simulated bottom hole pressures are very low in producers,
    or very high injectors.
    ---

    * **`ensembles`:** Which ensembles in `shared_settings` to visualize.
    * **`rel_file_pattern`:** path to `.arrow` files with summary data.
    ---
    Data is read directly from the arrow files with the raw frequency (not resampled).
    Resampling and csvs are not supported to avoid potential of interpolation, which
    might cover extreme BHP values.

    """

    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        ensembles: list,
        rel_file_pattern: str = "share/results/unsmry/*.arrow",
        wells: List[str] = None,
    ):
        super().__init__()

        self.ens_paths: Dict[str, Path] = {
            ensemble_name: webviz_settings.shared_settings["scratch_ensembles"][
                ensemble_name
            ]
            for ensemble_name in ensembles
        }

        self._input_provider_set = create_lazy_provider_set_from_paths(
            self.ens_paths,
            rel_file_pattern,
        )

        dfs = []
        column_keys = {}
        for ens_name in ensembles:
            ens_provider = self._input_provider_set.provider(ens_name)
            column_keys[ens_name] = _get_wbhp_vectors(ens_provider, wells)
            df = ens_provider.get_vectors_df(column_keys[ens_name], None)
            df["ENSEMBLE"] = ens_name
            dfs.append(df.loc[:, (df != 0).any(axis=0)])  # remove zero-columns

        self.smry = pd.concat(dfs)
        self.theme = webviz_settings.theme
        self.set_callbacks(app)

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.uuid("layout"),
                "content": (
                    "Dashboard for BHP QC:"
                    "Check that simulated bottom hole pressures are realistic."
                ),
            },
            {"id": self.uuid("ensemble"), "content": "Select ensemble to QC."},
            {
                "id": self.uuid("sort_by"),
                "content": "Sort wells left to right according to this value.",
            },
            {
                "id": self.uuid("n_wells"),
                "content": (
                    "Show max selected number of top ranked wells after sorting and filtering."
                ),
            },
            {"id": self.uuid("wells"), "content": "Filter wells."},
        ]

    @property
    def ensembles(self) -> List[str]:
        return list(self.smry["ENSEMBLE"].unique())

    @property
    def wells(self) -> List[set]:
        return sorted(
            list(set(col[5:] for col in self.smry.columns if col.startswith("WBHP:")))
        )

    @property
    def ens_colors(self) -> dict:
        return unique_colors(self.ensembles, self.theme)

    @property
    def label_map(self) -> Dict[str, str]:
        return {
            "Mean": "mean",
            "Count (data points)": "count",
            "Stddev": "std",
            "Minimum": "min",
            "Maximum": "max",
            "P10 (high)": "high_p10",
            "P50": "p50",
            "P90 (low)": "low_p90",
        }

    @property
    def layout(self) -> wcc.FlexBox:
        return wcc.FlexBox(
            id=self.uuid("layout"),
            children=[
                wcc.Frame(
                    style={"flex": 1, "height": "65vh"},
                    children=[
                        wcc.Dropdown(
                            label="Ensemble",
                            id=self.uuid("ensemble"),
                            options=[{"label": i, "value": i} for i in self.ensembles],
                            value=self.ensembles[0],
                            clearable=False,
                            multi=False,
                        ),
                        wcc.Dropdown(
                            label="Plot type",
                            id=self.uuid("plot_type"),
                            options=[
                                {"label": i, "value": i}
                                for i in [
                                    "Fan chart",
                                    "Bar chart",
                                    "Line chart",
                                ]
                            ],
                            clearable=False,
                            value="Fan chart",
                        ),
                        html.Div(
                            id=self.uuid("select_stat"),
                            style={"display": "none"},
                            children=[
                                wcc.SelectWithLabel(
                                    label="Select statistics",
                                    id=self.uuid("stat_bars"),
                                    options=[
                                        {"label": key, "value": value}
                                        for key, value in self.label_map.items()
                                    ],
                                    size=8,
                                    value=["min", "low_p90", "mean", "high_p10", "max"],
                                ),
                            ],
                        ),
                        wcc.Dropdown(
                            label="Sort by",
                            id=self.uuid("sort_by"),
                            options=[
                                {"label": key, "value": value}
                                for key, value in self.label_map.items()
                            ],
                            clearable=False,
                            value="low_p90",
                        ),
                        wcc.RadioItems(
                            vertical=False,
                            id=self.uuid("ascending"),
                            options=[
                                {"label": "Ascending", "value": True},
                                {"label": "Descending", "value": False},
                            ],
                            value=True,
                            labelStyle={"display": "inline-block"},
                        ),
                        wcc.Slider(
                            label="Max number of wells in plot",
                            id=self.uuid("n_wells"),
                            min=1,
                            max=len(self.wells),
                            step=1,
                            value=min(10, len(self.wells)),
                            marks={1: 1, len(self.wells): len(self.wells)},
                        ),
                        wcc.SelectWithLabel(
                            label="Wells",
                            id=self.uuid("wells"),
                            options=[{"label": i, "value": i} for i in self.wells],
                            size=min([len(self.wells), 20]),
                            value=self.wells,
                        ),
                    ],
                ),
                wcc.Frame(
                    color="white",
                    highlight=False,
                    style={"flex": 6, "height": "65vh"},
                    children=[
                        wcc.Graph(id=self.uuid("graph"), style={"height": "60vh"}),
                    ],
                ),
            ],
        )

    def set_callbacks(self, app: Dash) -> None:
        @app.callback(
            Output(self.uuid("graph"), "figure"),
            Input(self.uuid("ensemble"), "value"),
            Input(self.uuid("plot_type"), "value"),
            Input(self.uuid("n_wells"), "value"),
            Input(self.uuid("wells"), "value"),
            Input(self.uuid("sort_by"), "value"),
            Input(self.uuid("stat_bars"), "value"),
            Input(self.uuid("ascending"), "value"),
        )
        def _update_graph(
            ensemble: str,
            plot_type: str,
            n_wells: int,
            wells: Union[str, List[str]],
            sort_by: str,
            stat_bars: Union[str, List[str]],
            ascending: bool,
        ) -> dict:
            wells = wells if isinstance(wells, list) else [wells]
            stat_bars = stat_bars if isinstance(stat_bars, list) else [stat_bars]
            df = filter_df(df=self.smry, ensemble=ensemble, wells=wells)
            stat_df = (
                calc_statistics(df)
                .sort_values(sort_by, ascending=ascending)
                .iloc[0:n_wells, :]
            )
            traces = []
            if plot_type == "Fan chart":
                traces.extend(
                    _get_fanchart_traces(
                        ens_stat_df=stat_df,
                        color=self.ens_colors[ensemble],
                        legend_group=ensemble,
                    )
                )
            elif plot_type in ["Bar chart", "Line chart"]:
                for stat in stat_bars:
                    yaxis = "y2" if stat == "count" else "y"

                    if plot_type == "Bar chart":

                        traces.append(
                            {
                                "x": [vec[5:] for vec in stat_df.index],  # strip WBHP:
                                "y": stat_df[stat],
                                "name": [
                                    key
                                    for key, value in self.label_map.items()
                                    if value == stat
                                ][0],
                                "yaxis": yaxis,
                                "type": "bar",
                                "offsetgroup": stat,
                                "showlegend": True,
                            }
                        )
                    elif plot_type == "Line chart":
                        traces.append(
                            {
                                "x": [vec[5:] for vec in stat_df.index],  # strip WBHP:
                                "y": stat_df[stat],
                                "name": [
                                    key
                                    for key, value in self.label_map.items()
                                    if value == stat
                                ][0],
                                "yaxis": yaxis,
                                "type": "line",
                                "offsetgroup": stat,
                                "showlegend": True,
                            }
                        )
                    else:
                        raise ValueError("Invalid plot type.")

            layout = self.theme.create_themed_layout(
                {
                    "yaxis": {
                        "side": "left",
                        "title": "Bottom hole pressure",
                        "showgrid": True,
                    },
                    "yaxis2": {
                        "side": "right",
                        "overlaying": "y",
                        "title": "Count (data points)",
                        "showgrid": False,
                    },
                    "xaxis": {"showgrid": False},
                    "barmode": "group",
                    "legend": {"x": 1.05},
                }
            )
            return {"data": traces, "layout": layout}

        @app.callback(
            Output(self.uuid("select_stat"), "style"),
            Input(self.uuid("plot_type"), "value"),
        )
        def _update_stat_selector(plot_type: str) -> dict:
            return (
                {"display": "none"}
                if plot_type == "Fan chart"
                else {"display": "block"}
            )


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def filter_df(df: pd.DataFrame, ensemble: str, wells: List[str]) -> pd.DataFrame:
    """Filter dataframe for current ensembles and wells.
    Replacing zeroes (well not open) with np.NaN to not be accounted for
    in statistics.
    """
    columns = ["ENSEMBLE"] + [f"WBHP:{well}" for well in wells]
    return df.loc[df["ENSEMBLE"] == ensemble][columns].replace(0, np.NaN)


def calc_statistics(df: pd.DataFrame) -> pd.DataFrame:
    """Calculate statistics for given vectors over the ensembles
    refaxis is used if another column than DATE should be used to groupby.
    """
    # Invert p10 and p90 due to oil industry convention.
    def p10(x: np.ndarray) -> float:
        return np.nanpercentile(x, q=90)

    def p50(x: np.ndarray) -> float:
        return np.nanpercentile(x, q=50)

    def p90(x: np.ndarray) -> float:
        return np.nanpercentile(x, q=10)

    # Calculate statistics, ignoring NaNs.
    stat_df = (
        df.groupby("ENSEMBLE")
        .agg([np.nanmean, "count", np.nanstd, np.nanmin, np.nanmax, p10, p50, p90])
        .reset_index(drop=True, level="ENSEMBLE")
    )
    # Rename nanmin, nanmax and nanmean to min, max and mean.
    col_stat_label_map = {
        "nanmin": "min",
        "nanmax": "max",
        "nanmean": "mean",
        "nanstd": "std",
        "p10": "high_p10",
        "p90": "low_p90",
    }
    stat_df.rename(columns=col_stat_label_map, level=1, inplace=True)
    stat_df = stat_df.transpose().unstack()
    stat_df.columns = stat_df.columns.get_level_values(1)  # Remove 0 index column
    return stat_df


def _get_fanchart_traces(
    ens_stat_df: pd.DataFrame,
    color: str,
    legend_group: str,
) -> List[dict]:
    """Renders a fanchart for an ensemble vector"""

    x = [vec[5:] for vec in ens_stat_df.index]

    data = FanchartData(
        samples=x,
        low_high=LowHighData(
            low_data=ens_stat_df["low_p90"].values,
            low_name="P90 (low)",
            high_data=ens_stat_df["high_p10"].values,
            high_name="P10 (high)",
        ),
        minimum_maximum=MinMaxData(
            minimum=ens_stat_df["min"].values,
            maximum=ens_stat_df["max"].values,
        ),
        free_line=FreeLineData("Mean", ens_stat_df["mean"].values),
    )

    return get_fanchart_traces(
        data=data,
        hex_color=color,
        legend_group=legend_group,
        hovermode="x",
    )


# ---------------------------
def _get_wbhp_vectors(
    ens_provider: EnsembleSummaryProvider,
    wells: List[str] = None,
) -> list:
    """Return list of WBHP vectors. If wells arg is None, return for all wells."""

    if wells is not None:
        return [f"WBHP:{well}" for well in wells]

    wbhp_vectors = [
        vector for vector in ens_provider.vector_names() if vector.startswith("WBHP:")
    ]
    if not wbhp_vectors:
        raise RuntimeError("No WBHP vectors found.")

    return wbhp_vectors



from webviz_config.utils import StrEnum


class TreeModeOptions(StrEnum):
    STATISTICS = "statistics"
    SINGLE_REAL = "single_real"


class StatOptions(StrEnum):
    MEAN = "mean"
    P10 = "p10"
    P50 = "p50"
    P90 = "p90"
    MAX = "max"
    MIN = "min"


class NodeType(StrEnum):
    PROD = "prod"
    INJ = "inj"
    OTHER = "other"



import hashlib
import io
import json
import logging
import math
from dataclasses import asdict, dataclass
from typing import List, Optional, Tuple, Union
from urllib.parse import quote
from uuid import uuid4

import flask
import flask_caching
import xtgeo
from dash import Dash
from webviz_config.webviz_instance_info import WEBVIZ_INSTANCE_INFO

from webviz_subsurface._utils.perf_timer import PerfTimer

from ._surface_to_image import surface_to_png_bytes_optimized
from .ensemble_surface_provider import (
    ObservedSurfaceAddress,
    SimulatedSurfaceAddress,
    StatisticalSurfaceAddress,
    SurfaceAddress,
)

LOGGER = logging.getLogger(__name__)

_ROOT_URL_PATH = "/SurfaceServer"

_SURFACE_SERVER_INSTANCE: Optional["SurfaceServer"] = None


@dataclass(frozen=True)
class QualifiedSurfaceAddress:
    provider_id: str
    address: SurfaceAddress


@dataclass(frozen=True)
class QualifiedDiffSurfaceAddress:
    provider_id_a: str
    address_a: SurfaceAddress
    provider_id_b: str
    address_b: SurfaceAddress


@dataclass(frozen=True)
class SurfaceMeta:
    x_min: float
    x_max: float
    y_min: float
    y_max: float
    val_min: float
    val_max: float
    deckgl_bounds: List[float]
    deckgl_rot_deg: float  # Around upper left corner


class SurfaceServer:
    def __init__(self, app: Dash) -> None:
        cache_dir = (
            WEBVIZ_INSTANCE_INFO.storage_folder / f"SurfaceServer_filecache_{uuid4()}"
        )
        LOGGER.debug(f"Setting up file cache in: {cache_dir}")
        self._image_cache = flask_caching.Cache(
            config={
                "CACHE_TYPE": "FileSystemCache",
                "CACHE_DIR": cache_dir,
                "CACHE_DEFAULT_TIMEOUT": 0,
            }
        )
        self._image_cache.init_app(app.server)

        self._setup_url_rule(app)

    @staticmethod
    def instance(app: Dash) -> "SurfaceServer":
        # pylint: disable=global-statement
        global _SURFACE_SERVER_INSTANCE
        if not _SURFACE_SERVER_INSTANCE:
            LOGGER.debug("Initializing SurfaceServer instance")
            _SURFACE_SERVER_INSTANCE = SurfaceServer(app)

        return _SURFACE_SERVER_INSTANCE

    def publish_surface(
        self,
        qualified_address: Union[QualifiedSurfaceAddress, QualifiedDiffSurfaceAddress],
        surface: xtgeo.RegularSurface,
    ) -> None:
        timer = PerfTimer()

        if isinstance(qualified_address, QualifiedSurfaceAddress):
            base_cache_key = _address_to_str(
                qualified_address.provider_id, qualified_address.address
            )
        else:
            base_cache_key = _diff_address_to_str(
                qualified_address.provider_id_a,
                qualified_address.address_a,
                qualified_address.provider_id_b,
                qualified_address.address_b,
            )

        LOGGER.debug(
            f"Publishing surface (dim={surface.dimensions}, #cells={surface.ncol*surface.nrow}), "
            f"[base_cache_key={base_cache_key}]"
        )

        self._create_and_store_image_in_cache(base_cache_key, surface)

        LOGGER.debug(f"Surface published in: {timer.elapsed_s():.2f}s")

    def get_surface_metadata(
        self,
        qualified_address: Union[QualifiedSurfaceAddress, QualifiedDiffSurfaceAddress],
    ) -> Optional[SurfaceMeta]:

        if isinstance(qualified_address, QualifiedSurfaceAddress):
            base_cache_key = _address_to_str(
                qualified_address.provider_id, qualified_address.address
            )
        else:
            base_cache_key = _diff_address_to_str(
                qualified_address.provider_id_a,
                qualified_address.address_a,
                qualified_address.provider_id_b,
                qualified_address.address_b,
            )

        meta_cache_key = "META:" + base_cache_key
        meta: Optional[SurfaceMeta] = self._image_cache.get(meta_cache_key)
        if not meta:
            return None

        if not isinstance(meta, SurfaceMeta):
            LOGGER.error("Error loading SurfaceMeta from cache")
            return None

        return meta

    @staticmethod
    def encode_partial_url(
        qualified_address: Union[QualifiedSurfaceAddress, QualifiedDiffSurfaceAddress],
    ) -> str:

        if isinstance(qualified_address, QualifiedSurfaceAddress):
            address_str = _address_to_str(
                qualified_address.provider_id, qualified_address.address
            )
        else:
            address_str = _diff_address_to_str(
                qualified_address.provider_id_a,
                qualified_address.address_a,
                qualified_address.provider_id_b,
                qualified_address.address_b,
            )

        url_path: str = f"{_ROOT_URL_PATH}/{quote(address_str)}"
        return url_path

    def _setup_url_rule(self, app: Dash) -> None:
        @app.server.route(_ROOT_URL_PATH + "/<full_surf_address_str>")
        def _handle_surface_request(full_surf_address_str: str) -> flask.Response:
            LOGGER.debug(
                f"Handling surface_request: "
                f"full_surf_address_str={full_surf_address_str} "
            )

            timer = PerfTimer()

            img_cache_key = "IMG:" + full_surf_address_str
            LOGGER.debug(f"Looking for image in cache (key={img_cache_key}")

            cached_img_bytes = self._image_cache.get(img_cache_key)
            if not cached_img_bytes:
                LOGGER.error(
                    f"Error getting image for address: {full_surf_address_str}"
                )
                flask.abort(404)

            response = flask.send_file(
                io.BytesIO(cached_img_bytes), mimetype="image/png"
            )
            LOGGER.debug(
                f"Request handled from image cache in: {timer.elapsed_s():.2f}s"
            )
            return response

    def _create_and_store_image_in_cache(
        self,
        base_cache_key: str,
        surface: xtgeo.RegularSurface,
    ) -> None:

        timer = PerfTimer()
        LOGGER.debug("Converting surface to PNG image...")
        png_bytes: bytes = surface_to_png_bytes_optimized(surface)
        LOGGER.debug(f"Got PNG image, size={(len(png_bytes) / (1024 * 1024)):.2f}MB")
        et_to_image_s = timer.lap_s()

        img_cache_key = "IMG:" + base_cache_key
        meta_cache_key = "META:" + base_cache_key

        self._image_cache.add(img_cache_key, png_bytes)

        # For debugging rotations
        # unrot_surf = surface.copy()
        # unrot_surf.unrotate()
        # unrot_surf.quickplot("/home/sigurdp/gitRoot/hk-webviz-subsurface/quickplot.png")

        deckgl_bounds, deckgl_rot = _calc_map_component_bounds_and_rot(surface)

        meta = SurfaceMeta(
            x_min=surface.xmin,
            x_max=surface.xmax,
            y_min=surface.ymin,
            y_max=surface.ymax,
            val_min=surface.values.min(),
            val_max=surface.values.max(),
            deckgl_bounds=deckgl_bounds,
            deckgl_rot_deg=deckgl_rot,
        )
        self._image_cache.add(meta_cache_key, meta)
        et_write_cache_s = timer.lap_s()

        LOGGER.debug(
            f"Created image and wrote to cache in in: {timer.elapsed_s():.2f}s ("
            f"to_image={et_to_image_s:.2f}s, write_cache={et_write_cache_s:.2f}s), "
            f"[base_cache_key={base_cache_key}]"
        )


def _address_to_str(
    provider_id: str,
    address: SurfaceAddress,
) -> str:
    if isinstance(address, StatisticalSurfaceAddress):
        addr_type_str = "sta"
    elif isinstance(address, SimulatedSurfaceAddress):
        addr_type_str = "sim"
    elif isinstance(address, ObservedSurfaceAddress):
        addr_type_str = "obs"

    addr_hash = hashlib.md5(  # nosec
        json.dumps(asdict(address), sort_keys=True).encode()
    ).hexdigest()

    return f"{provider_id}___{addr_type_str}___{address.name}___{address.attribute}___{addr_hash}"


def _diff_address_to_str(
    provider_id_a: str,
    address_a: SurfaceAddress,
    provider_id_b: str,
    address_b: SurfaceAddress,
) -> str:
    return (
        "diff~~~"
        + _address_to_str(provider_id_a, address_a)
        + "~~~"
        + _address_to_str(provider_id_b, address_b)
    )


def _calc_map_component_bounds_and_rot(
    surface: xtgeo.RegularSurface,
) -> Tuple[List[float], float]:
    surf_corners = surface.get_map_xycorners()
    rptx = surf_corners[2][0]
    rpty = surf_corners[2][1]
    min_x = math.inf
    max_x = -math.inf
    min_y = math.inf
    max_y = -math.inf
    angle = -surface.rotation * math.pi / 180
    for coord in surf_corners:
        xpos = coord[0]
        ypos = coord[1]
        x_rotated = (
            rptx + ((xpos - rptx) * math.cos(angle)) - ((ypos - rpty) * math.sin(angle))
        )
        y_rotated = (
            rpty + ((xpos - rptx) * math.sin(angle)) + ((ypos - rpty) * math.cos(angle))
        )
        min_x = min(min_x, x_rotated)
        max_x = max(max_x, x_rotated)
        min_y = min(min_y, y_rotated)
        max_y = max(max_y, y_rotated)

    bounds = [
        min_x,
        min_y,
        max_x,
        max_y,
    ]

    return bounds, surface.rotation



from .ensemble_surface_provider import (
    EnsembleSurfaceProvider,
    ObservedSurfaceAddress,
    SimulatedSurfaceAddress,
    StatisticalSurfaceAddress,
    SurfaceAddress,
)
from .ensemble_surface_provider_factory import EnsembleSurfaceProviderFactory
from .surface_server import (
    QualifiedDiffSurfaceAddress,
    QualifiedSurfaceAddress,
    SurfaceMeta,
    SurfaceServer,
)



########################################
#
#  Copyright (C) 2021-     Equinor ASA
#
#  webviz-subsurface is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  webviz-subsurface is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http:#www.gnu.org/licenses/gpl.html>
#  for more details.
#
########################################
########################################
#
#  The code in this file is based on and/or inspired by opm-common,
#  which is distributed under the GNU General Public License v3.0,
#  and available at https://github.com/OPM/opm-common.
#
#  Especially in the following points this file, or files imported in this file
#  within the same context, make use of or are based on methods developed
#  for the opm-common project:
#  - Reading of Eclipse INIT files
#  - Units, unit systems and conversion of units
#  - Class structures for storing and accessing PVT data,
#    both base classes and derived classes for specific fluids
#
#  You should have received a copy of the GNU General Public License
#  along with webviz-subsurface. If not, see (https://www.gnu.org/licenses/gpl-3.0.html).
#
#  (12-01-2021)
#
########################################

import abc
import warnings
from enum import Enum
from typing import Any, Callable, List, Optional, Tuple, Union

import numpy as np
from opm.io.ecl import EclFile
from scipy import interpolate

from ..eclipse_unit import ConvertUnits, EclUnitEnum, EclUnits
from ..units import Unit


class EclPropertyTableRawData:  # pylint: disable=too-few-public-methods
    """
    A structure for storing read
    INIT file data.
    """

    def __init__(self) -> None:
        self.data = np.zeros(0)
        self.primary_key: List[int] = []
        self.num_primary = 0
        self.num_rows = 0
        self.num_cols = 0
        self.num_tables = 0


class PvxOBase(abc.ABC):
    """A common base class for all fluids.

    Should be inherited by any new fluid in order to have
    a common interface.
    """

    @abc.abstractmethod
    def get_keys(self) -> np.ndarray:
        """Returns all primary keys.

        Base implementation, raises a NotImplementedError.
        """

    @abc.abstractmethod
    def get_independents(self) -> np.ndarray:
        """Returns all independents.

        Base implementation, raises a NotImplementedError.
        """

    @abc.abstractmethod
    def formation_volume_factor(
        self, ratio: np.ndarray, pressure: np.ndarray
    ) -> np.ndarray:
        """Args:
            ratio: Ratio (key) values the volume factor values are requested for.
            pressure: Pressure values the volume factor values are requested for.

        Returns:
            All volume factor values for the given ratio and pressure values.

        Base implementation, raises a NotImplementedError.
        """

    @abc.abstractmethod
    def viscosity(self, ratio: np.ndarray, pressure: np.ndarray) -> np.ndarray:
        """Args:
            ratio: Ratio (key) values the viscosity values are requested for.
            pressure: Pressure values the viscosity values are requested for.

        Returns:
            All viscosity values for the given ratio and pressure values.

        Base implementation, raises a NotImplementedError.
        """

    @abc.abstractmethod
    def density(self, ratio: np.ndarray, pressure: np.ndarray) -> np.ndarray:
        """Args:
            ratio: Ratio (key) values the density values are requested for.
            pressure: Pressure values the density values are requested for.

        Returns:
            All density values for the given ratio and pressure values.

        Base implementation, raises a NotImplementedError.
        """


class PVxx(abc.ABC):
    """A base class for PVDx and PVTx"""

    @abc.abstractmethod
    def get_keys(self) -> np.ndarray:
        """Returns:
            All primary keys.

        Base implementation, raises a NotImplementedError.
        """

    @abc.abstractmethod
    def get_independents(self) -> np.ndarray:
        """Returns:
            All independents.

        Base implementation, raises a NotImplementedError.
        """

    @staticmethod
    def entry_valid(x: float) -> bool:
        """Returns:
        True if the given value is valid, i.e. < 1.0e20, else False.
        """
        return abs(x) < 1.0e20


class PVDx(PVxx):
    """A base class for dead and dry gas/oil respectively.

    Attributes:
        x: The independent values.
        y: A two-dimensional numpy array holding the dependent values.
        interpolation: A scipy interp1d object for interpolating the dependent values.
        inter_extrapolation: An extrap1d object for inter- and extrapolating the dependent values.
    """

    def __init__(
        self,
        index_table: int,
        raw: EclPropertyTableRawData,
        convert: ConvertUnits,
    ) -> None:
        """Extracts all values of the table with the given index from raw, converts them according
        to the given convert object and stores them as numpy arrays, x and y respectively.

        Creates an interpolation and an extrapolation object utilising scipy's interp1d and based on
        it the custom tailored extrap1d.

        Raises a ValueError if there is no interpolation interval given, that is when there are
        fewer than two independents.

        Args:
            index_table: The index of the table which values are supposed to be extracted.
            raw:
                An EclPropertyTableRawData object that was initialised based on an Eclipse
                INIT file.
            convert: A ConvertUnit object that contains callables for converting units.

        """
        self.x: np.ndarray = np.zeros(0)
        self.y: np.ndarray = np.zeros((raw.num_cols - 1, 0))

        column_stride = raw.num_rows * raw.num_tables * raw.num_primary
        table_stride = raw.num_primary * raw.num_rows

        for index_primary in range(0, raw.num_primary):
            if self.entry_valid(raw.primary_key[index_primary]):
                for index_row in range(0, raw.num_rows):
                    current_stride = (
                        index_table * table_stride
                        + index_primary * raw.num_rows
                        + index_row
                    )

                    if self.entry_valid(raw.data[column_stride * 0 + current_stride]):
                        self.x = np.append(
                            self.x,
                            convert.independent(
                                raw.data[column_stride * 0 + current_stride]
                            ),
                        )

                        self.y = np.append(
                            self.y,
                            [
                                [
                                    convert.column[index_column - 1](
                                        raw.data[
                                            column_stride * index_column
                                            + current_stride
                                        ]
                                    )
                                ]
                                for index_column in range(1, raw.num_cols)
                            ],
                            axis=1,
                        )

                    else:
                        break
            else:
                break

        if len(self.x) < 2:
            raise ValueError("No interpolation interval of non-zero size.")

        self.__interpolation = interpolate.interp1d(self.x, self.y, axis=1)

    def get_keys(self) -> np.ndarray:
        """Returns all primary keys.

        Since this is dry/dead gas/oil, there is no dependency on Rv/Rs.
        Hence, this method returns a list holding floats of value 0.0
        for each independent value.

        """
        return np.zeros(len(self.x))

    def get_independents(self) -> np.ndarray:
        """Returns all independents.

        In case of gas/oil this returns pressure values.

        """
        return self.x

    def formation_volume_factor(self, pressure: np.ndarray) -> np.ndarray:
        """Computes all formation volume factor values
        for the given pressure values.

        Args:
            pressure: Pressure values the volume factors are requested for.

        Returns:
            All formation volume factor values corresponding
            to the given pressure values.

        """
        # 1 / (1 / B)
        return self.__compute_quantity(pressure, lambda p: 1.0 / self.__fvf_recip(p))

    def viscosity(self, pressure: np.ndarray) -> np.ndarray:
        """Computes all viscosity values for the given pressure values.

        Args:
            pressure: Pressure values the viscosity values are requested for.

        Returns:
            All viscosity values corresponding
            to the given of pressure values.

        """
        # (1 / B) / (1 / (B * mu)
        return self.__compute_quantity(
            pressure, lambda p: self.__fvf_recip(p) / self.__fvf_mu_recip(p)
        )

    @staticmethod
    def __compute_quantity(
        pressures: np.ndarray, evaluate: Callable[[Any], Any]
    ) -> np.ndarray:
        """Calls the given evaluate function with each of the
        given pressure values and returns the results.

        Args:
            pressures: Pressure values
            evaluate: Evaluation function

        Returns:
            Values resulting from evaluating the pressure values.

        """
        result = np.zeros(len(pressures))

        for i, pressure in enumerate(pressures):
            result[i] = float(evaluate(pressure))

        return result

    def __fvf_recip(self, point: float) -> float:
        """Computes (possibly inter-/extrapolates) the reciprocal of
        the formation volume factor for the given point.

        Args:
            point:
                The pressure point the formation volume factor
                is requested for.

        Returns:
            The requested reciprocal formation volume factor.

        """
        return float(self.__interpolation(point)[0])

    def __fvf_mu_recip(self, point: float) -> float:
        """Computes (possibly inter-/extrapolates) the reciprocal of
        the product of the formation volume factor and viscosity
        for the given point.

        Args:
            point:
                The pressure point the formation volume factor
                is requested for.

        Returns:
            The requested reciprocal product of the formation volume factor
            and the viscosity.

        """
        return float(self.__interpolation(point)[1])


class PVTx(PVxx):
    def __init__(
        self,
        index_table: int,
        raw: EclPropertyTableRawData,
        convert: Tuple[
            Callable[
                [
                    float,
                ],
                float,
            ],
            ConvertUnits,
        ],
    ) -> None:
        """Extracts all values of the table with the given index from raw, converts them according
        to the given convert object and stores them as numpy arrays, keys, x and y respectively.

        Creates an interpolation and an extrapolation object utilising scipy's interp2d.

        Args:
            index_table: The index of the table which values are supposed to be extracted.
            raw:
                An EclPropertyTableRawData object that was initialised based on an Eclipse
                INIT file.
            convert:
                A tuple consisting of a callable for converting the primary keys and a
                ConvertUnit object that contains callables for converting units.

        """
        self.keys: np.ndarray = np.zeros(0)
        self.x: np.ndarray = np.zeros(0)
        self.y: List[np.ndarray] = [np.zeros(0) for _ in range(1, raw.num_cols)]

        column_stride = raw.num_rows * raw.num_tables * raw.num_primary
        table_stride = raw.num_primary * raw.num_rows

        for index_primary in range(0, raw.num_primary):
            if self.entry_valid(raw.primary_key[index_primary]):
                for index_row in range(0, raw.num_rows):
                    current_stride = (
                        index_table * table_stride
                        + index_primary * raw.num_rows
                        + index_row
                    )

                    if self.entry_valid(raw.data[column_stride * 0 + current_stride]):
                        self.keys = np.append(
                            self.keys, convert[0](raw.primary_key[index_primary])
                        )
                        self.x = np.append(
                            self.x,
                            convert[1].independent(
                                raw.data[column_stride * 0 + current_stride]
                            ),
                        )

                        for index_column in range(1, raw.num_cols):
                            self.y[index_column - 1] = np.append(
                                self.y[index_column - 1],
                                convert[1].column[index_column - 1](
                                    raw.data[
                                        column_stride * index_column + current_stride
                                    ]
                                ),
                            )

                    else:
                        break
            else:
                break

        # NOTE: If there is only one primary key, interp2d cannot be used.
        # As a fallback, use interp1d and make sure that the primary key asked for in
        # any of the methods of this instance is the one stored in self.keys[0].
        # Extrapolation is not possible.

        self.__single_key = np.amax(self.keys) == np.amin(self.keys)

        if len(self.x) < 2:
            raise ValueError("No interpolation interval of non-zero size.")

        # Ignore warnings here for now since this seems to be a bug in scipy
        # see: https://github.com/scipy/scipy/issues/4138
        # TODO: check if the warning can be fixed.

        warnings.filterwarnings("ignore")

        self.__interpolants: List[Union[interpolate.interp2d, interpolate.interp1d]] = [
            (
                interpolate.interp1d(self.x, self.y[index_column])
                if self.__single_key
                else interpolate.interp2d(
                    x=self.keys, y=self.x, z=self.y[index_column], kind="linear"
                )
            )
            for index_column in range(raw.num_cols - 1)
        ]

        warnings.filterwarnings("default")

    def get_keys(self) -> np.ndarray:
        """Returns all primary keys."""
        return self.keys

    def get_independents(self) -> np.ndarray:
        """Returns all independents."""
        return self.x

    def key_valid(self, key: np.ndarray) -> None:
        if self.__single_key and not all(k == self.keys[0] for k in key):
            raise ValueError(
                "Impossible to perform requested inter-/extrapolation due to insufficient data."
            )

    def formation_volume_factor(self, key: np.ndarray, x: np.ndarray) -> np.ndarray:
        """Computes formation volume factor values
        for the given ratio and pressure values.

        Args:
            key: Primary key values the volume factors are requested for.
            x: Independents the volume factors are requested for.

        Returns:
            Formation volume factor values corresponding
            to the given primary key and independent values.

        Base implementation, raises a NotImplementedError.

        """

        self.key_valid(key)

        return self.__compute_quantity(
            key,
            x,
            lambda curve, point: self.__interpolants[0](point)
            if self.__single_key
            else self.__interpolants[0](curve, point),
            lambda recip_fvf: 1.0 / recip_fvf,
        )

    def viscosity(self, key: np.ndarray, x: np.ndarray) -> np.ndarray:
        """Computes viscosity values for the given ratio and pressure values.

        Args:
            key: Primary key values the viscosity values are requested for.
            x: Independents the viscosity values are requested for.

        Returns:
            Viscosity values corresponding
            to the given primary key and independent values.

        Base implementation, raises a NotImplementedError.

        """

        self.key_valid(key)

        return self.__compute_quantity(
            key,
            x,
            lambda curve, point: [
                self.__interpolants[0](point),
                self.__interpolants[1](point),
            ]
            if self.__single_key
            else [
                self.__interpolants[0](curve, point),
                self.__interpolants[1](curve, point),
            ],
            lambda dense_vector: dense_vector[0] / dense_vector[1],
        )

    @staticmethod
    def __compute_quantity(
        key: np.ndarray,
        x: np.ndarray,
        inner_function: Callable,
        outer_function: Callable,
    ) -> np.ndarray:
        """Calls the evaluate method with each of the values
        in the given primary keys and the given inner_function
        and returns the results after the outer_function
        has been applied on each them.

        Args:
            key: Primary key values the viscosity values are requested for.
            x: Independents the viscosity values are requested for.
            inner_function: Callable for extracting a dense vector.
            outer_function:
                Callable that uses the dense vector to compute
                the requested quantity.

        Returns:
            Result values

        """

        num_vals = len(key)
        results = np.zeros(num_vals)
        if len(x) != num_vals:
            raise ValueError(
                "Number of inner sampling points does not match number of outer sampling points."
            )

        for i in range(0, num_vals):
            quantity = inner_function(key[i], x[i])

            results[i] = float(outer_function(quantity))

        return results


class InitFileDefinitions:  # pylint: disable=too-few-public-methods
    """
    A namespace for constant definitions for
    reading Eclipse INIT files.
    """

    LOGIHEAD_KW = "LOGIHEAD"
    INTEHEAD_KW = "INTEHEAD"
    INTEHEAD_UNIT_INDEX = 2
    INTEHEAD_PHASE_INDEX = 14
    LOGIHEAD_RS_INDEX = 0
    LOGIHEAD_RV_INDEX = 1
    TABDIMS_IBPVTO_OFFSET_ITEM = 6
    TABDIMS_JBPVTO_OFFSET_ITEM = 7
    TABDIMS_NRPVTO_ITEM = 8
    TABDIMS_NPPVTO_ITEM = 9
    TABDIMS_NTPVTO_ITEM = 10
    TABDIMS_IBPVTW_OFFSET_ITEM = 11
    TABDIMS_NTPVTW_ITEM = 12
    TABDIMS_IBPVTG_OFFSET_ITEM = 13
    TABDIMS_JBPVTG_OFFSET_ITEM = 14
    TABDIMS_NRPVTG_ITEM = 15
    TABDIMS_NPPVTG_ITEM = 16
    TABDIMS_NTPVTG_ITEM = 17
    LOGIHEAD_CONSTANT_OILCOMPR_INDEX = 39 - 1

    TABDIMS_IBDENS_OFFSET_ITEM = 18
    TABDIMS_NTDENS_ITEM = 19


class EclPhaseIndex(Enum):
    """Enumerator holding the different phases according
    to Eclipse file conventions"""

    AQUA = 0
    LIQUID = 1
    VAPOUR = 2


def is_const_compr_index() -> int:
    """Convenient function for better readibility.

    Returns:
        An integer that states that the oil has constant compression
        according to Eclipse LOGIHEAD file conventions.
    """
    return InitFileDefinitions.LOGIHEAD_CONSTANT_OILCOMPR_INDEX


def surface_mass_density(
    ecl_file: EclFile, phase: EclPhaseIndex, keep_unit_system: bool = True
) -> np.ndarray:
    """Extracts the surface mass density from the given Eclipse file for the given phase.

    Args:
        ecl_file: The Eclipse file to extract data from
        phase: Fluid phase to extract data for

    Returns:
        Surface mass density values.

    """
    if phase is EclPhaseIndex.LIQUID:
        col = 0
    elif phase is EclPhaseIndex.AQUA:
        col = 1
    elif phase is EclPhaseIndex.VAPOUR:
        col = 2
    else:
        raise AttributeError("Phase must be Liquid, Water or Vapour.")

    tabdims = ecl_file.__getitem__("TABDIMS")
    tab = ecl_file.__getitem__("TAB")

    start = tabdims[InitFileDefinitions.TABDIMS_IBDENS_OFFSET_ITEM] - 1
    nreg = tabdims[InitFileDefinitions.TABDIMS_NTDENS_ITEM]

    rho = tab[start + nreg * (col + 0) : start + nreg * (col + 1)]

    intehead = ecl_file.__getitem__(InitFileDefinitions.INTEHEAD_KW)
    unit_system = EclUnits.create_unit_system(
        intehead[InitFileDefinitions.INTEHEAD_UNIT_INDEX]
    )

    if not keep_unit_system:
        rho = [Unit.Convert.from_(rho_i, unit_system.density().value) for rho_i in rho]

    return rho


class FluidImplementation(abc.ABC):
    """Base class for fluid implementations

    Holds a list of regions (one per PVT table).

    Attributes:
        keep_unit_system: True if the original unit system was kept
        original_unit_system: An ErtEclUnitEnum representing the original unit system
    """

    class InvalidArgument(Exception):
        """An exception for invalid arguments"""

        def __init__(self, message: str):
            self.message = message
            super().__init__(message)

    class InvalidType(Exception):
        """An exception for invalid types"""

        def __init__(self) -> None:
            super().__init__("Invalid type. Only live oil/wet gas/water supported.")

    def __init__(self, unit_system: int, keep_unit_system: bool = True) -> None:
        """Initializes a fluid object.

        Args:
            unit_system: The original unit system
            keep_unit_system:
                True if the original unit system shall be kept,
                False if units shall be converted to SI units.

        """
        self._regions: List[PvxOBase] = []
        self.keep_unit_system = keep_unit_system
        self.original_unit_system = unit_system

    def pvdx_unit_converter(self) -> Optional[ConvertUnits]:
        """Creates a pseudo ConvertUnits object for PVDx interpolants
        that keeps the old unit system.

        Returns:
            Pseudo ConvertUnits object that does not do any unit conversions.

        """
        if self.keep_unit_system:
            return ConvertUnits(
                lambda x: x,
                [
                    lambda x: x,
                    lambda x: x,
                    lambda x: x,
                    lambda x: x,
                ],
            )
        return None

    def pvtx_unit_converter(
        self,
    ) -> Optional[Tuple[Callable[[float,], float,], ConvertUnits]]:
        """Creates a tuple consisting of a callable and a pseudo ConvertUnits
        object for PVTx interpolants which both keep the old unit system.

        Returns:
            Tuple of callable and pseudo ConvertUnits object
            which do not do any unit conversions.

        """
        if self.keep_unit_system:
            return (
                lambda x: x,
                ConvertUnits(
                    lambda x: x,
                    [
                        lambda x: x,
                        lambda x: x,
                        lambda x: x,
                        lambda x: x,
                    ],
                ),
            )
        return None

    @staticmethod
    def make_interpolants_from_raw_data(
        raw: EclPropertyTableRawData,
        construct: Callable[[int, EclPropertyTableRawData], PvxOBase],
    ) -> List[PvxOBase]:
        """Creates a list of interpolants from raw Eclipse table data using
        the given construct callable.

        Args:
            raw: Raw Eclipse table data to create interpolants from
            construct: Callable to use when creating interpolants

        Returns:
            A list of created interpolants.
        """
        interpolants: List[PvxOBase] = []

        for table_index in range(0, raw.num_tables):
            interpolants.append(construct(table_index, raw))

        return interpolants

    def pressure_unit(self, latex: bool = False) -> str:
        """Args:
            latex: True if the unit symbol shall be returned as Latex, False if not.

        Returns:
            A string containing the unit symbol of pressure.

        """
        unit_system = EclUnits.create_unit_system(
            self.original_unit_system
            if self.keep_unit_system
            else EclUnitEnum.ECL_SI_UNITS
        )

        if latex:
            return rf"${unit_system.pressure().symbol}$"
        return f"{unit_system.pressure().symbol}"

    def formation_volume_factor(
        self, region_index: int, ratio: np.ndarray, pressure: np.ndarray
    ) -> np.ndarray:
        """Args:
            region_index: Index of the requested PVT region
            ratio: Ratio values the data is requested for
            pressure: Pressure values the data is requested for

        Returns:
            Formation volume factor values according to the given values.

        """
        self.validate_region_index(region_index)

        return self._regions[region_index].formation_volume_factor(ratio, pressure)

    @abc.abstractmethod
    def formation_volume_factor_unit(self, latex: bool = False) -> str:
        """Args:
            latex: True if the unit symbol shall be returned as LaTeX, False if not.

        Returns:
            The unit symbol of the formation volume factor.

        Raises a NotImplementedError when called on a base class object.

        """

    def viscosity(
        self, region_index: int, ratio: np.ndarray, pressure: np.ndarray
    ) -> np.ndarray:
        """Args:
            region_index: Index of the requested PVT region
            ratio: Ratio values the data is requested for
            pressure: Pressure values the data is requested for

        Returns:
            Viscosity values according to the given values.

        """
        self.validate_region_index(region_index)

        return self._regions[region_index].viscosity(ratio, pressure)

    def viscosity_unit(self, latex: bool = False) -> str:
        """Creates and returns the unit symbol of the viscosity.

        Args:
            latex: True if the unit symbol shall be returned as LaTeX, False if not.

        Returns:
            The unit symbol of the viscosity.

        """
        unit_system = EclUnits.create_unit_system(
            self.original_unit_system
            if self.keep_unit_system
            else EclUnitEnum.ECL_SI_UNITS
        )

        if latex:
            return rf"${unit_system.viscosity().symbol}$"
        return unit_system.viscosity().symbol

    def density(
        self, region_index: int, ratio: np.ndarray, pressure: np.ndarray
    ) -> np.ndarray:
        """Args:
            region_index: Index of the requested PVT region
            ratio: Ratio values the data is requested for
            pressure: Pressure values the data is requested for

        Returns:
            Density values according to the given values.

        """
        self.validate_region_index(region_index)

        return self._regions[region_index].density(ratio, pressure)

    def density_unit(self, latex: bool = False) -> str:
        """Args:
            latex: True if the unit symbol shall be returned as LaTeX, False if not.

        Returns:
            The unit symbol of the density.

        """
        unit_system = EclUnits.create_unit_system(
            self.original_unit_system
            if self.keep_unit_system
            else EclUnitEnum.ECL_SI_UNITS
        )

        if latex:
            return rf"${unit_system.density().symbol}$"
        return unit_system.density().symbol

    @abc.abstractmethod
    def ratio_unit(self, latex: bool = False) -> str:
        """Args:
            latex: True if the unit symbol shall be returned as LaTeX, False if not.

        Returns:
            A string containing the unit symbol of the phase ratio (e.g. Sm³/Sm³).

        Raises a NotImplementedError when called on a base class object.

        """

    def get_region(self, region_index: int) -> PvxOBase:
        """Validates and returns the region at the given region_index.

        Args:
            region_index: Index of the requested PVT region

        Returns:
            The fluid interpolant related to the given region index.

        """
        self.validate_region_index(region_index)

        return self._regions[region_index]

    def validate_region_index(self, region_index: int) -> None:
        """Validates the given region index by ensuring that it is not out of range.

        Args:
            region_index: Index of the region to validate

        Raises:
            TypeError if the region is invalid.

        """
        if region_index >= len(self.regions()):
            if len(self.regions()) == 0:
                raise TypeError(
                    f"No oil PVT interpolant available in region {region_index + 1}."
                )

            raise TypeError(
                f"Region index {region_index + 1} outside valid range 0"
                f", ..., {len(self._regions) - 1}."
            )

    def regions(self) -> List[PvxOBase]:
        """Returns:
        All interpolants (one interpolant per region)

        """
        return self._regions

    def range_ratio(self, region_index: int) -> Tuple[float, float]:
        """Gets the primary key (ratio) range of the PVT region with the given index.

        Args:
            region_index: Index of the requested PVT region

        Returns:
            A tuple containing the min and max primary key (ratio) values.

        """
        region = self.get_region(region_index)
        return (min(region.get_keys()), max(region.get_keys()))

    def range_independent(self, region_index: int) -> Tuple[float, float]:
        """Gets the independent variable range of the PVT region with the given index.

        Args:
            region_index: Index of the requested PVT region

        Returns:
            A tuple containing the min and max independent values.

        """
        region = self.get_region(region_index)
        return (min(region.get_independents()), max(region.get_independents()))



import datetime
from typing import Any, Dict, List, Optional, Set, Union

import plotly.graph_objects as go
import webviz_core_components as wcc
from dash import ALL, Input, Output, State, callback, callback_context, html
from dash.development.base_component import Component
from webviz_config import WebvizConfigTheme
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC, ViewABC

from ..._types import ChartType
from ..._utils import EnsembleWellAnalysisData
from ._utils import WellOverviewFigure, format_well_overview_figure
from ._view_element import WellOverviewViewElement


class WellOverviewSettings(SettingsGroupABC):
    class Ids(StrEnum):
        CHARTTYPE = "charttype"
        ENSEMBLES = "ensembles"
        RESPONSE = "response"
        ONLY_PRODUCTION_AFTER_DATE = "only-production-after-date"
        CHARTTYPE_SETTINGS = "charttype-settings"
        CHARTTYPE_CHECKLIST = "charttype-checklist"

    def __init__(self, data_models: Dict[str, EnsembleWellAnalysisData]) -> None:
        super().__init__("Settings")

        self._ensembles = list(data_models.keys())

        dates: Set[datetime.datetime] = set()
        for _, ens_data_model in data_models.items():
            dates = dates.union(ens_data_model.dates)
        self._sorted_dates: List[datetime.datetime] = sorted(list(dates))

    def layout(self) -> List[Component]:
        settings_id = self.register_component_unique_id(
            WellOverviewSettings.Ids.CHARTTYPE_SETTINGS
        )
        checklist_id = self.register_component_unique_id(
            WellOverviewSettings.Ids.CHARTTYPE_CHECKLIST
        )
        return [
            wcc.RadioItems(
                id=self.register_component_unique_id(
                    WellOverviewSettings.Ids.CHARTTYPE
                ),
                label="Chart type",
                options=[
                    {"label": "Bar chart", "value": ChartType.BAR},
                    {"label": "Pie chart", "value": ChartType.PIE},
                    {"label": "Stacked area chart", "value": ChartType.AREA},
                ],
                value=ChartType.BAR,
                vertical=True,
            ),
            wcc.Dropdown(
                label="Ensembles",
                id=self.register_component_unique_id(
                    WellOverviewSettings.Ids.ENSEMBLES
                ),
                options=[{"label": col, "value": col} for col in self._ensembles],
                value=self._ensembles,
                multi=True,
            ),
            wcc.Dropdown(
                label="Response",
                id=self.register_component_unique_id(WellOverviewSettings.Ids.RESPONSE),
                options=[
                    {"label": "Oil production", "value": "WOPT"},
                    {"label": "Gas production", "value": "WGPT"},
                    {"label": "Water production", "value": "WWPT"},
                ],
                value="WOPT",
                multi=False,
                clearable=False,
            ),
            wcc.Dropdown(
                label="Only Production after date",
                id=self.register_component_unique_id(
                    WellOverviewSettings.Ids.ONLY_PRODUCTION_AFTER_DATE
                ),
                options=[
                    {
                        "label": dte.strftime("%Y-%m-%d"),
                        "value": dte.strftime("%Y-%m-%d"),
                    }
                    for dte in self._sorted_dates
                ],
                multi=False,
            ),
            html.Div(
                children=[
                    html.Div(
                        id={"id": settings_id, "charttype": ChartType.BAR},
                        children=wcc.Checklist(
                            id={"id": checklist_id, "charttype": ChartType.BAR},
                            label="Layout options",
                            options=[
                                {"label": "Show legend", "value": "legend"},
                                {"label": "Overlay bars", "value": "overlay_bars"},
                                {
                                    "label": "Show prod as text",
                                    "value": "show_prod_text",
                                },
                                {
                                    "label": "White background",
                                    "value": "white_background",
                                },
                            ],
                            value=["legend"],
                        ),
                    ),
                    html.Div(
                        id={"id": settings_id, "charttype": ChartType.PIE},
                        children=wcc.Checklist(
                            id={"id": checklist_id, "charttype": ChartType.PIE},
                            label="Layout options",
                            options=[
                                {"label": "Show legend", "value": "legend"},
                                {
                                    "label": "Show prod as text",
                                    "value": "show_prod_text",
                                },
                            ],
                            value=[],
                        ),
                    ),
                    html.Div(
                        id={"id": settings_id, "charttype": ChartType.AREA},
                        children=wcc.Checklist(
                            id={"id": checklist_id, "charttype": ChartType.AREA},
                            label="Layout options",
                            options=[
                                {"label": "Show legend", "value": "legend"},
                                {
                                    "label": "White background",
                                    "value": "white_background",
                                },
                            ],
                            value=["legend"],
                        ),
                    ),
                ],
            ),
        ]

    def set_callbacks(self) -> None:
        @callback(
            Output(
                {
                    "id": self.component_unique_id(
                        WellOverviewSettings.Ids.CHARTTYPE_SETTINGS
                    ).to_string(),
                    "charttype": ALL,
                },
                "style",
            ),
            Input(
                self.component_unique_id(
                    WellOverviewSettings.Ids.CHARTTYPE
                ).to_string(),
                "value",
            ),
            State(
                {
                    "id": self.component_unique_id(
                        WellOverviewSettings.Ids.CHARTTYPE_SETTINGS
                    ).to_string(),
                    "charttype": ALL,
                },
                "id",
            ),
        )
        @callback_typecheck
        def _display_charttype_settings(
            chart_selected: ChartType, charttype_settings_ids: list
        ) -> list:
            """Display only the settings relevant for the currently selected chart type."""
            return [
                {"display": "block"}
                if settings_id["charttype"] == chart_selected
                else {"display": "none"}
                for settings_id in charttype_settings_ids
            ]


class WellOverviewFilters(SettingsGroupABC):
    class Ids(StrEnum):
        SELECTED_WELLS = "selected-wells"

    def __init__(self, data_models: Dict[str, EnsembleWellAnalysisData]) -> None:
        super().__init__("Filters")
        self._wells: List[str] = []
        for ens_data_model in data_models.values():
            self._wells.extend(
                [well for well in ens_data_model.wells if well not in self._wells]
            )

    def layout(self) -> List[Component]:
        return [
            wcc.SelectWithLabel(
                label="Well",
                size=min(10, len(self._wells)),
                id=self.register_component_unique_id(
                    WellOverviewFilters.Ids.SELECTED_WELLS
                ),
                options=[{"label": well, "value": well} for well in self._wells],
                value=self._wells,
                multi=True,
            )
        ]


class WellOverviewView(ViewABC):
    class Ids(StrEnum):
        SETTINGS = "settings"
        FILTERS = "filters"
        CURRENT_FIGURE = "current-figure"
        VIEW_ELEMENT = "view-element"

    def __init__(
        self,
        data_models: Dict[str, EnsembleWellAnalysisData],
        theme: WebvizConfigTheme,
    ) -> None:
        super().__init__("Well overview")

        self._data_models = data_models
        self._theme = theme

        self.add_settings_group(
            WellOverviewSettings(self._data_models), self.Ids.SETTINGS
        )
        self.add_settings_group(
            WellOverviewFilters(self._data_models), self.Ids.FILTERS
        )
        main_column = self.add_column()
        main_column.add_view_element(WellOverviewViewElement(), self.Ids.VIEW_ELEMENT)

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.view_element(self.Ids.VIEW_ELEMENT)
                .component_unique_id(WellOverviewViewElement.Ids.GRAPH)
                .to_string(),
                "figure",
            ),
            Input(
                self.settings_group(self.Ids.SETTINGS)
                .component_unique_id(WellOverviewSettings.Ids.ENSEMBLES)
                .to_string(),
                "value",
            ),
            Input(
                {
                    "id": self.settings_group(self.Ids.SETTINGS)
                    .component_unique_id(WellOverviewSettings.Ids.CHARTTYPE_CHECKLIST)
                    .to_string(),
                    "charttype": ALL,
                },
                "value",
            ),
            Input(
                self.settings_group(self.Ids.SETTINGS)
                .component_unique_id(WellOverviewSettings.Ids.RESPONSE)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(self.Ids.SETTINGS)
                .component_unique_id(
                    WellOverviewSettings.Ids.ONLY_PRODUCTION_AFTER_DATE
                )
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(self.Ids.SETTINGS)
                .component_unique_id(WellOverviewSettings.Ids.CHARTTYPE)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(self.Ids.FILTERS)
                .component_unique_id(WellOverviewFilters.Ids.SELECTED_WELLS)
                .to_string(),
                "value",
            ),
            State(
                {
                    "id": self.settings_group(self.Ids.SETTINGS)
                    .component_unique_id(WellOverviewSettings.Ids.CHARTTYPE_CHECKLIST)
                    .to_string(),
                    "charttype": ALL,
                },
                "id",
            ),
            State(
                self.view_element(self.Ids.VIEW_ELEMENT)
                .component_unique_id(WellOverviewViewElement.Ids.GRAPH)
                .to_string(),
                "figure",
            ),
        )
        @callback_typecheck
        def _update_graph(
            ensembles: List[str],
            checklist_values: List[List[str]],
            sumvec: str,
            prod_after_date: Union[str, None],
            charttype_selected: ChartType,
            wells_selected: List[str],
            checklist_ids: List[Dict[str, str]],
            current_fig_dict: Optional[Dict[str, Any]],
        ) -> Component:
            """Updates the well overview graph with selected input (f.ex chart type)"""
            ctx = callback_context.triggered[0]["prop_id"].split(".")[0]

            settings = {
                checklist_id["charttype"]: checklist_values[i]
                for i, checklist_id in enumerate(checklist_ids)
            }

            # If the event is a plot settings event, then we only update the formatting
            # and not the figure data
            if (
                current_fig_dict is not None
                and self.settings_group(self.Ids.SETTINGS)
                .component_unique_id(WellOverviewSettings.Ids.CHARTTYPE_CHECKLIST)
                .to_string()
                in ctx
            ):
                fig_dict = format_well_overview_figure(
                    go.Figure(current_fig_dict),
                    charttype_selected,
                    settings[charttype_selected],
                    sumvec,
                    prod_after_date,
                )
            else:
                figure = WellOverviewFigure(
                    ensembles,
                    self._data_models,
                    sumvec,
                    datetime.datetime.strptime(prod_after_date, "%Y-%m-%d")
                    if prod_after_date is not None
                    else None,
                    charttype_selected,
                    wells_selected,
                    self._theme,
                )

                fig_dict = format_well_overview_figure(
                    figure.figure,
                    charttype_selected,
                    settings[charttype_selected],
                    sumvec,
                    prod_after_date,
                )

            return fig_dict



from pathlib import Path
from typing import Callable, Dict, List, Tuple

import pandas as pd
from webviz_config.webviz_store import webvizstore

from webviz_subsurface._datainput.fmu_input import scratch_ensemble


class GruptreeModel:
    """Facilitates loading of gruptree tables. Can be reused in all
    plugins that are using grouptree data and extended with additional
    functionality and filtering options if necessary.
    """

    def __init__(
        self,
        ens_name: str,
        ens_path: Path,
        gruptree_file: str,
        remove_gruptree_if_branprop: bool = True,
    ):
        self._ens_name = ens_name
        self._ens_path = ens_path
        self._gruptree_file = gruptree_file
        self._remove_gruptree_if_branprop = remove_gruptree_if_branprop
        self._dataframe = self.read_ensemble_gruptree()
        self._gruptrees_are_equal_over_reals = (
            self._dataframe["REAL"].nunique() == 1
            if not self._dataframe.empty
            else False
        )

    @property
    def dataframe(self) -> pd.DataFrame:
        """Returns a dataframe that will have the following columns:
        * DATE
        * CHILD (node in tree)
        * PARENT (node in tree)
        * KEYWORD (GRUPTREE, WELSPECS or BRANPROP)
        * REAL

        If gruptrees are exactly equal in all realizations then only one tree is
        stored in the dataframe. That means the REAL column will only have one unique value.
        If not, all trees are stored.
        """
        return self._dataframe

    @property
    def gruptrees_are_equal_over_reals(self) -> bool:
        """Returns true if gruptrees are exactly equal in all realizations."""
        return self._gruptrees_are_equal_over_reals

    def __repr__(self) -> str:
        """This is necessary for webvizstore to work on objects"""
        return f"""
GruptreeDataModel({self._ens_name!r}, {self._ens_path!r}, {self._gruptree_file!r})
        """

    @property
    def webviz_store(self) -> Tuple[Callable, List[Dict]]:
        return (
            self.read_ensemble_gruptree,
            [
                {
                    "self": self,
                }
            ],
        )

    @webvizstore
    def read_ensemble_gruptree(self) -> pd.DataFrame:
        """Reads the gruptree files for an ensemble from the scratch disk. These
        files can be exported in the FMU workflow using the ECL2CSV
        forward model with subcommand gruptree.

        If BRANPROP is found in the KEYWORD column, then GRUPTREE rows
        are filtered out.

        If the trees are equal in every realization, only one realization is kept.
        """

        ens = scratch_ensemble(self._ens_name, self._ens_path, filter_file="OK")
        df_files = ens.find_files(self._gruptree_file)

        if df_files.empty:
            return pd.DataFrame()

        # Load all gruptree dataframes and check if they are equal
        compare_columns = ["DATE", "CHILD", "KEYWORD", "PARENT"]
        df_prev = pd.DataFrame()
        dataframes = []
        gruptrees_are_equal = True
        for i, row in df_files.iterrows():
            df_real = pd.read_csv(row["FULLPATH"])

            if (
                self._remove_gruptree_if_branprop
                and "BRANPROP" in df_real["KEYWORD"].unique()
            ):
                df_real = df_real[df_real["KEYWORD"] != "GRUPTREE"]

            if (
                i > 0
                and gruptrees_are_equal
                and not df_real[compare_columns].equals(df_prev)
            ):
                gruptrees_are_equal = False
            else:
                df_prev = df_real[compare_columns].copy()

            df_real["REAL"] = row["REAL"]
            dataframes.append(df_real)
        df = pd.concat(dataframes)

        # Return either one or all realization in a common dataframe
        if gruptrees_are_equal:
            df = df[df["REAL"] == df["REAL"].min()]

        df["DATE"] = pd.to_datetime(df["DATE"])

        return df.where(pd.notnull(df), None)



import abc
from typing import List, Optional, Sequence

import pandas as pd


class EnsembleTableProvider(abc.ABC):
    @abc.abstractmethod
    def column_names(self) -> List[str]:
        ...

    @abc.abstractmethod
    def realizations(self) -> List[int]:
        ...

    @abc.abstractmethod
    def get_column_data(
        self, column_names: Sequence[str], realizations: Optional[Sequence[int]] = None
    ) -> pd.DataFrame:
        ...
